#charset "us-ascii"

/*
 *   Copyright 2003-04 Soren J. Lovborg.
 *  
 *   Permission is hereby granted, free of charge, to any person obtaining a
 *   copy of this software and associated documentation files (the "Software"), 
 *   to deal in the Software without restriction, including without limitation 
 *   the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *   and/or sell copies of the Software, and to permit persons to whom the
 *   Software is furnished to do so, subject to the following conditions:
 *  
 *   The above copyright notice and this permission notice shall be included in
 *   all copies or substantial portions of the Software.
 *  
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 *
 */

#include <tads.h>
#define false nil

/* ------------------------------------------------------------------------ */

class CompilerOperator: CompilerFunctionCall
    op = nil
    needsNegation = false
    funcName = (op.funcName)
    
    construct(op_)
    {
        if (op_.negatedOp) { op = op_.negatedOp; needsNegation = true; }
        else op = op_;

        values = new Vector(2);
    }

    addPart(op_, value)
    {
        local x;
        local finalop_ = op_.negatedOp ? op_.negatedOp : op_;
        
        if (op != finalop_ || !op.multipParam && values.length() >= 2)
        {
            /*
             * We have to add a new call.
             * Create a new function, add ourself as first param.
             */
             
            x = createInstance(op_);
            x.add(self);
        }
        else x = self;
        
        x.add(value);
        return x;
    }

    prepareCode()
    {
        local x = inherited();
        if (!needsNegation) return x;

        needsNegation = false;
        local y = new CompilerOperator(opNot);
        y.add(x);
        return y.prepareCode();
    }

    constantString = (op.calculateString(values))

    optimize()
    {
        local x = inherited();
        local str = x.constantString;
        return str ? CompilerConstant.createFromString(str) : x;
    }
;

/* ------------------------------------------------------------------------ */

class Operator: Token
    funcName = nil
    negatedOp = nil
    token = nil
    
    getCompilerValue = (new CompilerOperator(self))
;

class UnaryOperator: Operator
    calculateString(vals)
    {
        local x = vals[1].constantNumber;
        if (x == nil) return nil;

        x = calc(x);
        if (x == nil) return nil;

        return toString(x);
    }
;

class BinaryOperator: Operator
    multiParam = false

    calculateString(vals)
    {
        local x = vals[1].constantNumber;
        if (x == nil) return nil;
        
        for (local i = 2; i <= vals.length(); i++)
        {
            local cv = vals[i].constantNumber();
            if (cv == nil) return nil;

            x = calc(x, cv);
            if (x == nil) return nil;
        }
        return toString(x);
    }
;

Operator template 'token' 'funcName';
Operator template 'token' !negatedOp;

/* ------------------------------------------------------------------------ */

opNot:          UnaryOperator  '!'     '@not'
    calc(x) { return x == 1 ? 0 : 1; }
;

opNeg:          UnaryOperator
    funcName = '@neg'
    calc(x) { return -x; }
;

opConcat:       BinaryOperator '&'     '@print'  multipParam = true
    getCompilerValue = (new CompilerPrintCall())
;

opPlus:         BinaryOperator '+'     '@add' multipParam = true
    calc(a, b) { return a + b; }
;
opMinus:        BinaryOperator '-'     '@sub'
    calc(a, b) { return a - b; }
;
opMultiply:     BinaryOperator '*'     '@mul' multipParam = true
    calc(a, b) { return a * b; }
;
opDivide:       BinaryOperator '/'     '@div'
    calc(a, b) { return a / b; }
;
opIDivide:      BinaryOperator, WordToken 'div'   '@idiv'
    calc(a, b) { return a % b; }
;
opModulos:      BinaryOperator, WordToken 'mod'   '@mod'
    calc(a, b) { return a % b; }
;
opEqual:        BinaryOperator '=='    '@eq'  multipParam = true
    calculateString(vals)
    {
        local x = vals[1].constantString();
        if (x == nil || x.find('%')) return nil;
        for (local i = 2; i <= vals.length(); i++)
        {
            local cv = vals[i].constantString();
            if (cv == nil || cv.find('%')) return nil;
            if (x != cv) return '0';
        }
        return '1';
    }
;
opLess:         BinaryOperator '<'     '@lt'
    calc(a, b) { return a < b ? 1 : 0; }
;
opGreater:      BinaryOperator '>'     '@gt'
    calc(a, b) { return a > b ? 1 : 0; }
;
opAnd:          BinaryOperator, WordToken 'and'    '@and' multipParam = true
    calc(a, b) { return a && b ? 1 : 0; }
;
opOr:           BinaryOperator, WordToken 'or'    '@or'   multipParam = true
    calc(a, b) { return a || b ? 1 : 0; }
;
opXor:          BinaryOperator, WordToken 'xor'   '@xor'
    calc(a, b) { return (a == 1) != (b == 1) ? 1 : 0; }
;
opHas:          BinaryOperator, WordToken 'has'   '@testflag';

opNotEqual:     BinaryOperator '!='    !opEqual;
opLessEqual:    BinaryOperator '<='    !opGreater;
opGreaterEqual: BinaryOperator '>='    !opLess;
opHasnt:        BinaryOperator, WordToken 'hasnt' !opHas;

/* ------------------------------------------------------------------------ */
